/*******************************************************************************
 * 
 * jKiwi -- A virtual makeover and hairstyler application
 * Copyright (c) 2007-2008 Dan Mihai Ile, Maria Joao Roque Barbado Leal
 * 
 * This file is part of jKiwi.
 * 
 * jKiwi is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 * 
 * jKiwi is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 *******************************************************************************/

package tabs;

import java.awt.AlphaComposite;
import java.awt.BasicStroke;
import java.awt.Graphics2D;
import java.awt.Point;
import java.awt.Rectangle;
import java.awt.Stroke;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.awt.event.MouseMotionListener;
import java.awt.geom.AffineTransform;
import java.awt.image.AffineTransformOp;
import java.awt.image.BufferedImage;
import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.ResourceBundle;

import javax.imageio.ImageIO;


import org.eclipse.swt.SWT;
import org.eclipse.swt.events.DisposeEvent;
import org.eclipse.swt.events.DisposeListener;
import org.eclipse.swt.events.SelectionAdapter;
import org.eclipse.swt.events.SelectionEvent;
import org.eclipse.swt.graphics.Color;
import org.eclipse.swt.graphics.Image;
import org.eclipse.swt.layout.GridData;
import org.eclipse.swt.layout.GridLayout;
import org.eclipse.swt.widgets.Button;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Control;
import org.eclipse.swt.widgets.Display;
import org.eclipse.swt.widgets.Event;
import org.eclipse.swt.widgets.Label;
import org.eclipse.swt.widgets.Listener;
import org.eclipse.swt.widgets.Scale;

import core.Loader;

import blending.BlendingComposite;

import utils.ImageUtils;
import utils.StockIcon;

/**
 * Instances of this class represents a tab that
 * can be used to do blush to a person's face
 *
 */
public class BlushTab extends TabContainer {

	/**
	 * Instances of this class represents a state of the {@link BlushTab}
	 */
	private class BlushHistoryStep {

		/**
		 * The left cheek coordinates
		 */
		private final Rectangle leftCheekCoord;

		/**
		 * The right cheek coordinates
		 */
		private final Rectangle rightCheekCoord;

		/**
		 * The selected button
		 */
		private final ColorButton selectedColorButton;

		/**
		 * The opacity value of the blush effect
		 */
		private final int cheekOpacityScaleValue;

		/**
		 * The cheek image size
		 */
		private final int cheekSizeScaleValue;

		/**
		 * Creates a new instance of this class to save the
		 * {@link BlushTab} given state
		 * @param leftCheekCoord the left cheek coordinates
		 * @param rightCheekCoord the right cheek coordinates
		 * @param colorButton the selected button
		 * @param cheekOpacityScaleValue the opacity value of the blush effect
		 * @param cheekSizeScaleValue the cheek image size
		 */
		public BlushHistoryStep(Rectangle leftCheekCoord, Rectangle rightCheekCoord, ColorButton colorButton, int cheekOpacityScaleValue, int cheekSizeScaleValue) {

			this.leftCheekCoord = new Rectangle(leftCheekCoord);
			this.rightCheekCoord = new Rectangle(rightCheekCoord);

			//search for the reference of the actual button instead of the reference of the selectedButton
			ColorButton selectedColorButton = null;
			if (colorButton != null) {
				for (Control c : colorButton.getParent().getChildren()) {
					if (c instanceof ColorButton) {
						ColorButton cb = (ColorButton) c;
						if (cb.equals(colorButton)) {
							selectedColorButton = cb;
							break;
						}
					}
				}
			}
			this.selectedColorButton = selectedColorButton;
			this.cheekOpacityScaleValue = cheekOpacityScaleValue;
			this.cheekSizeScaleValue = cheekSizeScaleValue;
		}
	}

	/**
	 * The opacity that will be used to draw the blush
	 */
	private static float ALPHA_COMPOSITE_OPACITY = 0.4f;

	/**
	 * Alpha composite used to draw the blush image
	 */
	private AlphaComposite currentCompositeOpacity;

	/**
	 * The opacity of the blush image bounding box
	 */
	private static float BOUNDS_OPACITY = 0.2f;

	/**
	 * Alpha composite used to draw the blush image bounding box
	 */
	private AlphaComposite boundsCompositeOpacity;

	/**
	 * messages that are to be used by the entire GUI
	 */
	private ResourceBundle messages;

	/**
	 * A list with all the colors used by ColorButton
	 * For each color in this list one ColorButton will be created
	 */
	private Color[] colorsList = {
			new Color(Display.getCurrent(), 244, 188, 128),
			new Color(Display.getCurrent(), 220, 169, 115),
			new Color(Display.getCurrent(), 193, 137, 77),
			new Color(Display.getCurrent(), 195, 128, 57),
			new Color(Display.getCurrent(), 198, 118, 33),
			new Color(Display.getCurrent(), 203, 107, 3),
			new Color(Display.getCurrent(), 255, 184, 159),
			new Color(Display.getCurrent(), 237, 156, 127),
			new Color(Display.getCurrent(), 215, 143, 117),
			new Color(Display.getCurrent(), 200, 136, 114),
			new Color(Display.getCurrent(), 169, 116, 98),
			new Color(Display.getCurrent(), 132, 90, 76),
			new Color(Display.getCurrent(), 244, 167, 161),
			new Color(Display.getCurrent(), 246, 147, 140),
			new Color(Display.getCurrent(), 232, 121, 113),
			new Color(Display.getCurrent(), 200, 104, 96),
			new Color(Display.getCurrent(), 176, 90, 84),
			new Color(Display.getCurrent(), 157, 80, 74),
			new Color(Display.getCurrent(), 255, 142, 171),
			new Color(Display.getCurrent(), 255, 101, 141),
			new Color(Display.getCurrent(), 241, 57, 104),
			new Color(Display.getCurrent(), 213, 48, 90),
			new Color(Display.getCurrent(), 188, 40, 78),
			new Color(Display.getCurrent(), 160, 31, 64),
			new Color(Display.getCurrent(), 255, 193, 200),
			new Color(Display.getCurrent(), 243, 175, 183),
			new Color(Display.getCurrent(), 225, 161, 169),
			new Color(Display.getCurrent(), 208, 148, 155),
			new Color(Display.getCurrent(), 193, 135, 142),
			new Color(Display.getCurrent(), 156, 108, 114)
	};

	//the left and right cheek blush images
	private BufferedImage blushLeftCheek = null;
	private BufferedImage blushRightCheek = null;

	//the left and right coordinates of the cheek images
	private Rectangle leftCheekCoord = null;
	private Rectangle rightCheekCoord = null;

	/**
	 * Used to set the upper and color buttons composite
	 */
	private TabGuiUtility tabGui;

	/**
	 * the tab's main composite
	 */
	private Composite mainComposite;

	/**
	 * All the colored buttons in use by this tab
	 */
	private ArrayList<ColorButton> coloredButtonList;

	/**
	 * The blush original image file read from the file system
	 */
	private BufferedImage blushFile;

	//mouse listeners in the drawing area
	private MouseListener mouseListener;
	private MouseMotionListener mouseMotionListener;

	//flags to indicate if the mouse is over the left or right cheek
	protected boolean mouseOverLeftCheek;
	protected boolean mouseOverRightCheek;

	//used to help with the dragging of the blush images over the drawing area
	protected int startX;
	protected int startY;
	protected Point oldLeftCheekPoint;
	protected Point oldRightCheekPoint;

	/**
	 * Composite for the bottom controls
	 */
	private Composite bottomComposite;

	//bottom composite label separators
	private Label labelSeparatorUp;
	private Label labelSeparatorDown;

	//Scale for the size and opacity of the blush images
	private Scale sizeScale;
	private Scale opacityScale;

	//labels for the size and opacity scale
	private Label labelSizeScale;
	private Label labelOpacityScale;

	/**
	 * Button for clearing this tab's contents on the drawing area
	 */
	private Button clearButton;

	/**
	 * flag to indicate whether the tab has or not access to call for redrawing the image area
	 */
	protected boolean imageAreaRedrawAllowed;

	/**
	 * A reference to this instance class
	 */
	private BlushTab thisBlushTab = this;

	/**
	 * Constructs a new instance of this class given its parent,
	 * a style value describing its behavior and appearance and
	 * this instance's position in the {@link TabsContainer} array
	 *
	 * @param parent a {@link TabsContainer} which will be the parent of the new instance
	 * @param style the style of the tab
	 * @param tabLayerPosition the tab's position in the {@link TabsContainer} array
	 *
	 */
	public BlushTab(TabsContainer parent, int style, int tabLayerPosition) {
		super(parent, style, tabLayerPosition);

		messages = Loader.getMessages();

		//the 2 things needed to be done by every implementation of TabComposite
		super.setTabImage(new Image(Display.getCurrent(), System.getProperty("program.pixmapspath") + "tab_icon_blush.png"));
		super.setTabToolTipText(messages.getString("BLUSH_TAB_TOOLTIP"));

		addDisposeListener(new DisposeListener() {
			public void widgetDisposed(DisposeEvent e) {
				for (Color c : colorsList) {
					c.dispose();
				}
			}
		});

		imageAreaRedrawAllowed = true;

		//create the tab's GUI
		tabGui = new TabGuiUtility(this);
		mainComposite = this;
		tabGui.createUpperComposite(super.getTabToolTipText());
		coloredButtonList = tabGui.createColorButtonsComposite(colorsList);
		tabGui.setSelectedColorButton(null);
		createBottomComposite();

		//default opacity
		currentCompositeOpacity = AlphaComposite.getInstance(AlphaComposite.SRC_OVER, ALPHA_COMPOSITE_OPACITY);
		boundsCompositeOpacity = AlphaComposite.getInstance(AlphaComposite.SRC_OVER, BOUNDS_OPACITY);

		//read the blush image file
		createColoredBlushImage();

		extendColorButtonsBehavior();

		leftCheekCoord = new Rectangle(95, 250, blushFile.getWidth(), blushFile.getHeight());
		rightCheekCoord = new Rectangle(205, 250, blushFile.getWidth(), blushFile.getHeight());

	}

	/**
	 * This method initializes bottomComposite
	 */
	public void createBottomComposite() {

		//grid data for the clear button
		GridData gridDataClearButton = new GridData();
		gridDataClearButton.grabExcessHorizontalSpace = true;
		gridDataClearButton.horizontalSpan = 2;
		gridDataClearButton.horizontalAlignment = org.eclipse.swt.layout.GridData.FILL;

		//grid data for the separator label up
		GridData gridDataSeparatorLabelUp = new GridData();
		gridDataSeparatorLabelUp.grabExcessHorizontalSpace = true;
		gridDataSeparatorLabelUp.horizontalSpan = 2;
		gridDataSeparatorLabelUp.horizontalAlignment = org.eclipse.swt.layout.GridData.FILL;

		//grid data for the separator label down
		GridData gridDataSeparatorLabelDown = new GridData();
		gridDataSeparatorLabelDown.horizontalAlignment = org.eclipse.swt.layout.GridData.FILL;
		gridDataSeparatorLabelDown.horizontalSpan = 2;
		gridDataSeparatorLabelDown.grabExcessHorizontalSpace = true;

		//grid data for the opacity scale
		GridData gridDataOpacityScale = new GridData();
		gridDataOpacityScale.grabExcessHorizontalSpace = true;
		gridDataOpacityScale.horizontalAlignment = org.eclipse.swt.layout.GridData.FILL;

		//grid data for the size scale
		GridData gridDataSizeScale = new GridData();
		gridDataSizeScale.grabExcessHorizontalSpace = true;
		gridDataSizeScale.horizontalAlignment = org.eclipse.swt.layout.GridData.FILL;

		//grid data for the composite
		GridData gridDataComposite = new GridData();
		gridDataComposite.verticalAlignment = org.eclipse.swt.layout.GridData.END;
		gridDataComposite.grabExcessHorizontalSpace = true;
		gridDataComposite.horizontalAlignment = org.eclipse.swt.layout.GridData.FILL;
		gridDataComposite.grabExcessVerticalSpace = true;

		//grid layout for the composite
		GridLayout gridLayoutComposite = new GridLayout();
		gridLayoutComposite.numColumns = 2;

		//initialize the composite
		bottomComposite = new Composite(mainComposite, SWT.NONE);
		bottomComposite.setLayout(gridLayoutComposite);
		bottomComposite.setLayoutData(gridDataComposite);

		//initialize the upper label separator
		labelSeparatorUp = new Label(bottomComposite, SWT.SEPARATOR | SWT.HORIZONTAL);
		labelSeparatorUp.setLayoutData(gridDataSeparatorLabelUp);

		//initialize the clear button
		clearButton = new Button(bottomComposite, SWT.NONE);
		clearButton.setText(messages.getString("BLUSH_TAB_CLEAR"));
		clearButton.setImage(new Image(Display.getCurrent(), StockIcon.getIcon(StockIcon.ICON_CLEAR, StockIcon.SIZE_MENU, true)));
		clearButton.setLayoutData(gridDataClearButton);

		//initialize the size scale label
		labelSizeScale = new Label(bottomComposite, SWT.NONE);
		labelSizeScale.setText(messages.getString("BLUSH_TAB_SIZE"));

		//initialize the size scale
		sizeScale = new Scale(bottomComposite, SWT.NONE);
		sizeScale.setSelection(50);
		sizeScale.setMinimum(20);
		sizeScale.setMaximum(80);
		sizeScale.setLayoutData(gridDataSizeScale);

		//initialize the opacity scale label
		labelOpacityScale = new Label(bottomComposite, SWT.NONE);
		labelOpacityScale.setText(messages.getString("BLUSH_TAB_OPACITY"));

		//initialize the opacity scale
		opacityScale = new Scale(bottomComposite, SWT.NONE);
		opacityScale.setSelection(50);
		opacityScale.setLayoutData(gridDataOpacityScale);

		//initialize the lower separator label
		labelSeparatorDown = new Label(bottomComposite, SWT.SEPARATOR | SWT.HORIZONTAL);
		labelSeparatorDown.setLayoutData(gridDataSeparatorLabelDown);

		//clear button listener
		clearButton.addSelectionListener(new SelectionAdapter() {
			public void widgetSelected(SelectionEvent e) {
				//TODO check if history step is created only when needed
				if (e.widget instanceof Button) {
					tabsContainer.createHistoryStep();
					clearImageArea();
					if (imageAreaRedrawAllowed) {
						tabsContainer.redrawImageArea(thisBlushTab);
					}
				}
			}
		});

		//opacity scale listeners
		opacityScale.addSelectionListener(new org.eclipse.swt.events.SelectionAdapter() {
			public void widgetSelected(org.eclipse.swt.events.SelectionEvent e) {
				double opacityInterval = 0.3;
				double adjustInterval = (((float)opacityScale.getSelection() / opacityScale.getMaximum()) * opacityInterval) - (opacityInterval / 2.0);
				currentCompositeOpacity = AlphaComposite.getInstance(AlphaComposite.SRC_OVER, (float)(ALPHA_COMPOSITE_OPACITY + adjustInterval));
				if (imageAreaRedrawAllowed) {
					tabsContainer.redrawImageArea(thisBlushTab);
				}
			}
		});
		opacityScale.addListener(SWT.MouseDown, new Listener() {
			public void handleEvent(Event event) {
				tabsContainer.createHistoryStep();
			}
		});

		//size scale listeners
		sizeScale.addSelectionListener(new org.eclipse.swt.events.SelectionAdapter() {
			public void widgetSelected(org.eclipse.swt.events.SelectionEvent e) {
				createColoredCheeks();
				if (imageAreaRedrawAllowed) {
					tabsContainer.redrawImageArea(thisBlushTab);
				}
			}
		});
		sizeScale.addListener(SWT.MouseDown, new Listener() {
			public void handleEvent(Event event) {
				tabsContainer.createHistoryStep();
			}
		});
	}

	/**
	 * Color buttons should do more than just updating the {@link ColorDifference}
	 * All color buttons listeners are implemented here
	 */
	private void extendColorButtonsBehavior() {
		for (final ColorButton button : coloredButtonList) {
			button.addMouseListener(new org.eclipse.swt.events.MouseAdapter() {

				public void mouseDown(org.eclipse.swt.events.MouseEvent e) {
					//TODO history step should be created only when the image changed
					tabsContainer.createHistoryStep();
					//trick the flags so that borders are shown when changing cheek colors
					mouseOverLeftCheek = true;
					mouseOverRightCheek = true;
					if (imageAreaRedrawAllowed) {
						tabsContainer.redrawImageArea(thisBlushTab);
					}
				}

				public void mouseUp(org.eclipse.swt.events.MouseEvent e) {
					try {
						Thread.sleep(100);
						createColoredCheeks();
						if (imageAreaRedrawAllowed) {
							tabsContainer.redrawImageArea(thisBlushTab);
						}
						Thread.sleep(100);
					} catch (InterruptedException e1) {
						// TODO Auto-generated catch block
						e1.printStackTrace();
					}
					//restore the flags values
					mouseOverLeftCheek = false;
					mouseOverRightCheek = false;
					createColoredCheeks();
					if (imageAreaRedrawAllowed) {
						tabsContainer.redrawImageArea(thisBlushTab);
					}
				}
			});
		}
	}

	/**
	 * Reads the blush image from the file system
	 * resizes it and blends it with the current selected color
	 * from the colored button
	 * @return true if image creation succeeded
	 */
	private boolean createColoredBlushImage() {

		double widthFactor = sizeScale.getSelection() / 50.0;
		double heightFactor = sizeScale.getSelection() / 50.0;
		AffineTransform tx = new AffineTransform(); //use affine transform class to do the resizing
		tx.scale(widthFactor, heightFactor);
		AffineTransformOp op = new AffineTransformOp(tx, AffineTransformOp.TYPE_BICUBIC);

		try {
			blushFile = ImageIO.read(new File(System.getProperty("program.pixmapspath") + "tab_blush_B3.png"));
			//TODO resize should keep the blush in the middle and not to extend to lower right
			blushFile = op.filter(blushFile, null);
			if (leftCheekCoord != null && rightCheekCoord != null) {
				leftCheekCoord.width = blushFile.getWidth();
				leftCheekCoord.height = blushFile.getHeight();
				rightCheekCoord.width = blushFile.getWidth();
				rightCheekCoord.height = blushFile.getHeight();				
			}
			//apply the color to the blush image
			if (tabGui.getCurrentColor() == null) return false;
			java.awt.Composite compositeToUse = Loader.getCompositeFor(BlendingComposite.COLOR_OVERLAY);
			blushFile = ImageUtils.applyBlending(blushFile, compositeToUse, tabGui.getCurrentColor().getRed(), tabGui.getCurrentColor().getGreen(), tabGui.getCurrentColor().getBlue());

		} catch (IOException e1) {
			// TODO Auto-generated catch block
			e1.printStackTrace();
			return false;
		}
		return true;
	}

	/**
	 * Clears this tab's image in the drawing area
	 * and resets the needed variables
	 */
	private void clearImageArea() {
		tabGui.setSelectedColorButton(null);
		blushLeftCheek = null;
		blushRightCheek = null;
	}

	/**
	 * Create the two cheeks images from the blush image file
	 */
	private void createColoredCheeks() {
		if (!createColoredBlushImage()) return ;

		//create the blush right cheek
		blushRightCheek = new BufferedImage(blushFile.getWidth(), blushFile.getHeight(), BufferedImage.TYPE_INT_ARGB);
		Graphics2D g2dRightCheek = blushRightCheek.createGraphics();
		g2dRightCheek.drawImage(blushFile, 0, 0, null);
		g2dRightCheek.dispose();

		//flip the blush file
		AffineTransform tx = AffineTransform.getScaleInstance(-1, 1);
		tx.translate(-blushFile.getWidth(null), 0);
		AffineTransformOp op = new AffineTransformOp(tx, AffineTransformOp.TYPE_NEAREST_NEIGHBOR);
		blushFile = op.filter(blushFile, null);

		//create the left cheek
		blushLeftCheek = new BufferedImage(blushFile.getWidth(), blushFile.getHeight(), BufferedImage.TYPE_INT_ARGB);
		Graphics2D g2dLeftCheek = blushLeftCheek.createGraphics();
		g2dLeftCheek.drawImage(blushFile, 0, 0, null);
		g2dLeftCheek.dispose();
	}

	@Override
	public BufferedImage getLayerImage() {
		//create a copy of this tab's image and return it
		BufferedImage transparentCheeksImage = null;
		if (blushLeftCheek != null && blushRightCheek != null) {
			transparentCheeksImage = new BufferedImage(tabsContainer.getOriginalImage().getWidth(), tabsContainer.getOriginalImage().getHeight(), BufferedImage.TYPE_INT_ARGB);
			Graphics2D g2d = transparentCheeksImage.createGraphics();
			drawCheeksImage(g2d);
			g2d.dispose();
		}
		return transparentCheeksImage;
	}

	/**
	 * Draws the cheek image using the given graphics context
	 * @param g2d the graphics context used for drawing operations
	 */
	private void drawCheeksImage(Graphics2D g2d) {

		//backup the current composite
		java.awt.Composite defaultComposite = g2d.getComposite();

		//set blush image opacity
		g2d.setComposite(currentCompositeOpacity);

		//draw the left and right cheeks
		g2d.drawImage(blushLeftCheek, leftCheekCoord.x, leftCheekCoord.y, null);
		g2d.drawImage(blushRightCheek, rightCheekCoord.x, rightCheekCoord.y, null);

		//restore the g2d original composite
		g2d.setComposite(defaultComposite);

		//check if mouse is in cheek area and highlight the area
		//TODO check if cursor is visible when drawing the rectangle because of full screen preview
		if (mouseOverLeftCheek) {
			Stroke defaultStroke = g2d.getStroke();
			g2d.setStroke(new BasicStroke(2,1,1));
			g2d.setComposite(boundsCompositeOpacity);
			g2d.drawRect(leftCheekCoord.x, leftCheekCoord.y, leftCheekCoord.width, leftCheekCoord.height);
			g2d.setStroke(defaultStroke);
			g2d.setComposite(defaultComposite);
		}
		if (mouseOverRightCheek) {
			Stroke defaultStroke = g2d.getStroke();
			g2d.setStroke(new BasicStroke(2,1,1));
			g2d.setComposite(boundsCompositeOpacity);
			g2d.drawRect(rightCheekCoord.x, rightCheekCoord.y, rightCheekCoord.width, rightCheekCoord.height);
			g2d.setStroke(defaultStroke);
			g2d.setComposite(defaultComposite);
		}
	}

	@Override
	public void drawCurrentLayer(Graphics2D g2d) {
		//if there exists left and right cheek draw them to the drawing area
		if (blushLeftCheek != null && blushRightCheek != null) {
			drawCheeksImage(g2d);
		}
	}

	@Override
	public void clear() {
		clearImageArea();
	}

	@Override
	public Object getCurrentState() {
		return new BlushHistoryStep(
				leftCheekCoord,
				rightCheekCoord,
				tabGui.getSelectedButton(),
				opacityScale.getSelection(),
				sizeScale.getSelection()
		);
	}

	@Override
	public void setCurrentState(Object o) {
		if (o instanceof BlushHistoryStep) {
			BlushHistoryStep hs = (BlushHistoryStep) o;

			imageAreaRedrawAllowed = false;

			tabGui.setSelectedColorButton(hs.selectedColorButton);
			leftCheekCoord = hs.leftCheekCoord;
			rightCheekCoord = hs.rightCheekCoord;

			opacityScale.setSelection(hs.cheekOpacityScaleValue);
//			opacityScale.notifyListeners(eventType, event) //TODO this is how should work instead of the 3 lines below 
			double opacityInterval = 0.3;
			double adjustInterval = (((float)opacityScale.getSelection() / opacityScale.getMaximum()) * opacityInterval) - (opacityInterval / 2.0);
			currentCompositeOpacity = AlphaComposite.getInstance(AlphaComposite.SRC_OVER, (float)(ALPHA_COMPOSITE_OPACITY + adjustInterval));

			sizeScale.setSelection(hs.cheekSizeScaleValue);
			createColoredCheeks();

			if (hs.selectedColorButton == null) {
				blushLeftCheek = null;
				blushRightCheek = null;
			}
			imageAreaRedrawAllowed = true;
		}
	}

	@Override
	public MouseListener mouseListener() {
		if (mouseListener == null) {
			mouseListener = new MouseListener() {

				//unimplemented methods
				public void mouseClicked(MouseEvent e) {}
				public void mouseExited(MouseEvent e) {}

				public void mouseEntered(MouseEvent e) {
					e.getComponent().setCursor(new java.awt.Cursor(java.awt.Cursor.DEFAULT_CURSOR));
				}

				public void mousePressed(MouseEvent e) {
					if (mouseOverLeftCheek || mouseOverRightCheek) {

						//save the tab's current state before changing anything
						tabsContainer.getDisplay().syncExec(new Runnable() {
							public void run() {
								tabsContainer.createHistoryStep();
							}
						});

						//enable dragging for the left cheek
						if (mouseOverLeftCheek) {
							startX = e.getX();
							startY = e.getY();
							oldLeftCheekPoint = new Point(leftCheekCoord.x, leftCheekCoord.y);
							e.getComponent().setCursor(new java.awt.Cursor(java.awt.Cursor.MOVE_CURSOR));
						}

						//enable dragging for the right cheek
						if (mouseOverRightCheek) {
							startX = e.getX();
							startY = e.getY();
							oldRightCheekPoint = new Point(rightCheekCoord.x, rightCheekCoord.y);
							e.getComponent().setCursor(new java.awt.Cursor(java.awt.Cursor.MOVE_CURSOR));
						}
					}
				}

				public void mouseReleased(MouseEvent e) {
					if (mouseOverLeftCheek || mouseOverRightCheek) {
						e.getComponent().setCursor(new java.awt.Cursor(java.awt.Cursor.HAND_CURSOR));
					} else {
						e.getComponent().setCursor(new java.awt.Cursor(java.awt.Cursor.DEFAULT_CURSOR));
					}
				}
			};
		}
		return mouseListener;
	}

	@Override
	public MouseMotionListener mouseMotionListener() {

		if (mouseMotionListener == null) {
			mouseMotionListener = new MouseMotionListener() {

				public void mouseDragged(MouseEvent e) {
					if (mouseOverLeftCheek) {
						leftCheekCoord.x = e.getX() - startX + oldLeftCheekPoint.x;
						leftCheekCoord.y = e.getY() - startY + oldLeftCheekPoint.y;
					}
					if (mouseOverRightCheek) {
						rightCheekCoord.x = e.getX() - startX + oldRightCheekPoint.x;
						rightCheekCoord.y = e.getY() - startY + oldRightCheekPoint.y;
					}
					if (mouseOverLeftCheek || mouseOverRightCheek) {
						if (imageAreaRedrawAllowed) {
							tabsContainer.redrawImageArea(thisBlushTab);
						}
					}
				}

				public void mouseMoved(MouseEvent e) {

					if (blushLeftCheek == null || blushRightCheek == null) return ;

					//check if mouse is over cheek
					boolean oldLeft = mouseOverLeftCheek;
					boolean oldRight = mouseOverRightCheek;
					mouseOverLeftCheek = leftCheekCoord.contains(e.getPoint());
					mouseOverRightCheek = rightCheekCoord.contains(e.getPoint());
					if (mouseOverLeftCheek != oldLeft || mouseOverRightCheek != oldRight) {
						if (imageAreaRedrawAllowed) {
							tabsContainer.redrawImageArea(thisBlushTab);
						}
					}
					if (mouseOverLeftCheek || mouseOverRightCheek) {
						e.getComponent().setCursor(new java.awt.Cursor(java.awt.Cursor.HAND_CURSOR));
					} else {
						e.getComponent().setCursor(new java.awt.Cursor(java.awt.Cursor.DEFAULT_CURSOR));
					}
				}

			};
		}
		return mouseMotionListener;
	}

	public void originalImageChanged() {
		// TODO Auto-generated method stub
	}

	public void disposeData() {
		blushFile = null;
		blushLeftCheek = null;
		blushRightCheek = null;
	}
}